/*****************************************
 * 
 * 9/08/2016 - Chris P. 
 * Program to control the Mitsubishi Heavy Industries (MHI) Air conditioners uising a NodeMCU v1.0 ESP8266-12E
 * This script has been based on a script by Vincent Cruvellier url - https://github.com/r45635/HVAC-IR-Control/blob/master/HVAC_ESP8266/HVAC_ESP8266.ino
 * Which inturn is based on / uses the IRremote library for Arduino by Ken Shirriff - http://arcfn.com/2009/08/multi-protocol-infrared-remote-library.html
 * 
 ********Original comments form Vincent Cruevellier *************************************
 * The Sketch use an ESP 8266 to read DHT sensor Values, Send HVAC IR command and sleep.
 * 
 * I have used ESP 8266 ESP-12 Module, the IR led Emitter should be drive from a 5V current. You can power the IR LED from the 3.3V but the 
 * emission distance will be considerably reduced. The best is to drive the IR led from 5V and drive through 
 * a transitor the led emission (Signal). Caution a lot of IR-Led module do not have real Signal drive.
 * Indeed, lot of them have Signal directly connected to VDD.
 * 
 * Sketch Exemple with ESP8266 + HVAC IR Emission Capability + DHT Reading
 * Sketch has been very simplified to not use IR-Remote lib but just what we need for this example
 * ESP Deep sleep Feauture is used on this sketch. That requires ESP/RST connected to ESP/GPIO16 
 *   in order to wakeup from the deep sleep.
 * 
 * Hardware Connection
 *  IR LED SIGNAL => ESP/GPIO_4
 *  DHT SIGNAL => ESP/GPIO_5
 *  RESET => ESP/GPIO_16
 * 
 ****************************************/

#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include "DHT.h"

//WIFI Constants
const char* ssid = "<SSID>";            //  SSID of LAN/WLAN
const char* password = "<PASSWORD>";        //  password

//MQTT constants
const char* mqttServer = "192.168.0.10"; //TODO: replace with server name... 
const int mqttPort = 1883;
//const int led0 = 16; // Inbuild LED pin

//MQTT Topics
const char* mqttTopicHum = "hum";
const char* mqttTopicTemp = "temp";
const String mqttTopicMode = "mode";
const String mqttTopicSetTemp = "settemp";
const String mqttTopicFan = "fan";
const String mqttTopicOn = "on";
const String mqttTopicProfile = "profile";
const String mqttTopicGo = "go";

//Define and then instantiate the wifi / MQTT Client
WiFiClient client;
PubSubClient mqttClient(client);

//Defines for the DHT22 (AM2302)
#define DHTPIN 5     // what digital pin we're connected to
#define DHTTYPE DHT22   // DHT 22  (AM2302), AM2321

// Initialize DHT sensor.
DHT dht(DHTPIN, DHTTYPE);

//Sensor varables
float hum; //Hold humidity value 
float temp; // Hold the Temp value

//Global IR Declarations 
int halfPeriodicTime;
int IRpin;
int khz;

// HVAC MITSUBISHI Heavy Industries IR Mark & Space timings
#define HVAC_MITSUBISHI_HI_HDR_MARK    3368 
#define HVAC_MITSUBISHI_HI_HDR_SPACE   1684 
#define HVAC_MITSUBISHI_HI_BIT_MARK    421 
#define HVAC_MITSUBISHI_HI_ONE_SPACE   1263 
#define HVAC_MISTUBISHI_HI_ZERO_SPACE  421 
#define HVAC_MITSUBISHI_HI_RPT_MARK    421 
#define HVAC_MITSUBISHI_HI_RPT_SPACE   17100 // Above original iremote limit

long previousMillis = 0;
long interval = 5000;

byte data[19]; 

/* callback
 * function for reading incoming message
 */
void callback(char* topic, byte* payload, unsigned int length){

  Serial.print("Message arrived [");
  Serial.print(topic);
  Serial.print("] ");
  
  String sTopic(topic);
  String sMessy = (String)(char*)payload;
  String sMessage = sMessy.substring(0, length);
  

  Serial.println("Char array");
  Serial.println((char*)payload);
  Serial.println(length);
  Serial.print("Message: ");
  Serial.print(sMessage);
  
  //Update HVAC settings based on messages in sub-routines, unless Go message received then send the IR code. 
  if(sTopic == mqttTopicOn){setSwitch(sMessage); }
  else if (sTopic == mqttTopicTemp) {setTemp(sMessage); }
  else if (sTopic == mqttTopicFan) {setFan(sMessage); }
  else if (sTopic == mqttTopicProfile) {setProfile(sMessage); }
  else if (sTopic == mqttTopicMode) {setMode(sMessage); }
  else if (sTopic == mqttTopicGo){ sendHvacMitsubishiHI(); }
  else {}
}

/* setSwitch
 *  function to modify IR array bytes 6 & 7 for On/off
 */
void setSwitch(String sMessage){
  // Byte 6 & 7 - On / Off and mode
  if (sMessage.toInt() == 0) {
    //Switch Off in Auto mode
    data[5] = (byte) 0xFF; 
    data[6] = (byte) 0x00;
  } else {
    //Switch On in Auto mode
    data[5] = (byte) 0xF7; 
    data[6] = (byte) 0x08;
  }
  Serial.println("Post switch update: ");
  for (int i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
}

/* setMode
 * function to modify bytes 6 & 7 of the IR array to reflect the requested mode
 */
void setMode(String sMessage){
const String COLD = "Cold";

  Serial.println();
  Serial.println(sMessage);
  Serial.println();

  if (sMessage == "Auto"){data[5] = (byte) 0xF7; data[6] = (byte) 0x08; }
  else if (sMessage == "Cold"){ Serial.println("here"); data[5] = (byte) 0xF6; data[6] = (byte) 0x09; }
  else if (sMessage == "Dry"){  Serial.println("there");data[5] = (byte) 0xF5; data[6] = (byte) 0x0A; }
  else if (sMessage == "Fan"){  data[5] = (byte) 0xF4; data[6] = (byte) 0x0B; }
  else if (sMessage == "Hot"){   data[5] = (byte) 0xF3; data[6] = (byte) 0x0C; }
  else {Serial.println("problem....");};
  Serial.println();
  Serial.println("Post mode update: ");
  for (int i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
  Serial.println();
}

/*  setTemp
 *  function to modify bytes 8 & 9 of the IR array to reflect the requested temp
 */
void setTemp(String sMessage){
  // Byte 8 & 9 - Temperature
  // Check Min Max For Hot Mode
  int HVAC_Temp = sMessage.toInt();
  int iTemp;
  if (HVAC_Temp > 30) { iTemp = 30;}
  else if (HVAC_Temp < 18) { iTemp = 18; } 
  else { iTemp = HVAC_Temp; };
  
  data[7] = (byte) (255 - (iTemp - 17));
  data[8] = (byte) iTemp - 17;

  Serial.println("Post Temp update: ");
  for (int i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
  Serial.println();
  
}

/*  setFan
 *  function to modify bytes 10 & 11 of the IR array to reflect the requested fan setting
 */
void setFan(String sMessage){
    
  // Byte 10 & 11 - FAN
  switch (sMessage.toInt())
  {
    case 0:    data[9] = (byte) 0xFF; data[10] = (byte) 0x00; break; //Set to Auto
    case 1:       data[9] = (byte) 0xFE; data[10] = (byte) 0x01; break; //set to fan speed 1
    case 2:       data[9] = (byte) 0xFD; data[10] = (byte) 0x02; break; //set to fan speed 2
    case 3:       data[9] = (byte) 0xFC; data[10] = (byte) 0x03; break; //set to fan speed 3
    case 4:       data[9] = (byte) 0xFB; data[10] = (byte) 0x04; break; //set to fan speed 4
    default: break;
  }
  Serial.println();
  Serial.print("Post Fan update: ");
  for (int i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
  Serial.println();
}

/*  setProfile
 *  function to modify bytes 10 & 11 of the IR array to reflect the requested Profile setting(e.g. Power / Eco)setting
 *  This overwrites the Fan Setting
 */
void setProfile(String sMessage){
  // Profile to overrite - Byte 10 & 11
  if(sMessage == "NORMAL") { data[9] = (byte) 0xFF; data[10] = (byte) 0x00;} // set fan to auto.
  else if(sMessage == "ECO") {data[9] = (byte) 0xF9; data[10] = (byte) 0x06;}
  else if(sMessage == "POWER"){data[9] = (byte) 0xF3; data[10] = (byte) 0x0C;}
  else{}
  Serial.println("Post Profile update: ");
  for (int i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
}

/* reconnect
 * Function to connect the MQTT client using factoryish pattern
 */
void reconnect(){

  //loop until reconnected
  while (!mqttClient.connected()){
    Serial.print("Attempting MQTT connection...");
    //attempt to connect
    if (mqttClient.connect("ESP8266 Client")){
      Serial.println("connected");
      //Subscribe to the settings
      mqttClient.subscribe("on");
      mqttClient.subscribe("settemp");
      mqttClient.subscribe("fan");
      mqttClient.subscribe("profile");
      mqttClient.subscribe("mode");
      mqttClient.subscribe("go");
      
    } else {
      Serial.print("failed, rc=");
      Serial.print(mqttClient.state());
      Serial.println(" , try again in 5 seconds...");
      //wait... 
      delay(5000);
    }
  }
}

/*
 * setup
 * Function that runs on start up
 * Used to establish Valid IR Array.
 */
void setup() {

  //Set IR PIN and PIN mode
  IRpin=4;
  khz=38;
  halfPeriodicTime = 500/khz;
  pinMode(IRpin, OUTPUT);

  //Valid AC IR Frame -> On, Auto 23 degrees, and auto fan.
  //              52    AE    C3    1A    E5    F3    0C    F3    0C    FF    00    FF    00    FF    00    FF    00    7F    80

  data[0] = 0x52;
  data[1] = 0xAE;
  data[2] = 0xC3;
  data[3] = 0x1A;
  data[4] = 0xE5;
  data[5] = 0xF7;
  data[6] = 0x08;
  data[7] = 0xF9;
  data[8] = 0x06;
  data[9] = 0xFF;
  data[10] = 0x00;
  data[11] = 0xFF;
  data[12] = 0x00;
  data[13] = 0xFF;
  data[14] = 0x00;
  data[15] = 0xFF;
  data[16] = 0x00;
  data[17] = 0x7F;
  data[18] = 0x80;
  
  Serial.begin(115200);                        //  start serial for debug
  delay(10);
  
  //Start the Temp sensor library
  dht.begin();

  mqttClient.setServer(mqttServer, mqttPort);
  mqttClient.setCallback(callback);

  Serial.println("Let the games begin...");
}

/*
 * Continiuosly executes
 */
void loop() {

  //If no connection exists establish one.
  if (!mqttClient.connected()){
    reconnect();
  }
  mqttClient.loop();

  // Non-blocking delay to slow down the temp refreshes to avoud flooding the MQTT client. 
  // To be removed once the ESP sleep is in. 
  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis > interval){
    previousMillis = currentMillis;
    
    //Read values and publish to MQTT queues
    ReadSensor();
  
  }
  
  // Put the ESP NODEMCU to sleep to save power
  //Serial.println("bedtime");
  //ESP.deepSleep(60 * 1000000, WAKE_RF_DEFAULT);// the 1000000 changes to seconds >> deepSleep for 1 minute
  //Serial.println("Going to Sleep now");
  delay(500);
  //The ESP will go to sleep now and will send a timed signal via D0 (GPIO16) to RST
  //The ESP will then reset back into wake up mode
  
}

/*
 * ReadSensor
 * Reads the value of the temp sensor and sends temp and humidity strings to the MQTT server
 */
void ReadSensor(){

  //Read values form sensor
  hum = dht.readHumidity();
  temp = dht.readTemperature();

  //Publish values to MQTT queue
  if(!isnan(hum)){
    //PrintHumidity();
    PublishString("hum",String(hum));
  }
  if(!isnan(temp)){
    //PrintTemp();
    PublishString("temp",String(temp));
  }

}

/* PublishString
 * Helper function to publish a string message to an MQTT queue
 */
void PublishString(String sTopic, String sPayload){
  char cTopic[50];
  char cMessage[50];
  
  //Convert message to char array
  sTopic.toCharArray(cTopic, sTopic.length() + 1);
  sPayload.toCharArray(cMessage, sPayload.length() + 1);
  mqttClient.publish(cTopic,cMessage);
  
}

/*****************************************************************************************
** Function to Send IR commands to a Mitsubishi Heavy Industries Split system AC, 
** Remote model: RLA502A700B. Note IRAnalysIR reports that it is using Sanyo152AC.
*****************************************************************************************/
void sendHvacMitsubishiHI()
{

#define  HVAC_MITSUBISHI_DEBUG;  // Un comment to access DEBUG information through Serial Interface

byte mask = 1; //our bitmask
byte i;

#ifdef HVAC_MITSUBISHI_DEBUG
  Serial.println("Packet to send: ");
  for (i = 0; i < 19; i++) {
    Serial.print("_"); Serial.print(data[i], HEX);
  }
  Serial.println(".");
  for (i = 0; i < 19; i++) {
    Serial.print(data[i], BIN); Serial.print(" ");
  }
  Serial.println(".");
#endif

  enableIROut(38);  // 38khz
  space(0);
  
  // Header for the Packet
  mark(HVAC_MITSUBISHI_HI_HDR_MARK);
  space(HVAC_MITSUBISHI_HI_HDR_SPACE);
  
  //loop through each hex value in the array.   
  for (i = 0; i < 19; i++) {
    
    // Send all Bits from Byte Data
    mask = 00000001;        
    //iterate through the byte by bit. 
    for(int k=0; k<8; k++){

      //Compare the byte against the target bit using the mask. 
      if (data[i] & mask) { // Bit ONE
        //Serial.print("1");
        mark(HVAC_MITSUBISHI_HI_BIT_MARK);
        space(HVAC_MITSUBISHI_HI_ONE_SPACE);
      }
      else { // Bit ZERO
        //Serial.print("0");
        mark(HVAC_MITSUBISHI_HI_BIT_MARK);
        space(HVAC_MISTUBISHI_HI_ZERO_SPACE);
      }
      
      //Shift the mask one place to the right, adding a zero to the front of the number.
      mask <<= 1;
    }
  }
    
  // End of Packet and retransmission of the Packet
  mark(HVAC_MITSUBISHI_HI_RPT_MARK);
  space(HVAC_MITSUBISHI_HI_RPT_SPACE);
  space(0); // Just to be sure

}

/****************************************************************************
/* enableIROut : Set global Variable for Frequency IR Emission
/***************************************************************************/ 
void enableIROut(int khz) {
  // Enables IR output.  The khz value controls the modulation frequency in kilohertz.
  halfPeriodicTime = 500/khz; // T = 1/f but we need T/2 in microsecond and f is in kHz
}

/****************************************************************************
/* mark ( int time) 
/***************************************************************************/ 
void mark(int time) {
  // Sends an IR mark for the specified number of microseconds.
  // The mark output is modulated at the PWM frequency.
  long beginning = micros();
  while(micros() - beginning < time){
    digitalWrite(IRpin, HIGH);
    delayMicroseconds(halfPeriodicTime);
    digitalWrite(IRpin, LOW);
    delayMicroseconds(halfPeriodicTime); //38 kHz -> T = 26.31 microsec (periodic time), half of it is 13
  }
}

/****************************************************************************
/* space ( int time) 
/***************************************************************************/ 
/* Leave pin off for time (given in microseconds) */
void space(int time) {
  // Sends an IR space for the specified number of microseconds.
  // A space is no output, so the PWM output is disabled.
  digitalWrite(IRpin, LOW);
  if (time > 0) delayMicroseconds(time);
}

/*PrintHumidity
 * helper function to print humidity value to serial
 */
void PrintHumidity(){
  Serial.print("Humidity: ");
  Serial.print(hum);
  Serial.println("%");
}


/*PrintTemp
 * helper function to print temp value to serial
 */
void PrintTemp(){
  Serial.print("Temperature: "); 
  Serial.print(temp);
  Serial.println("C.");
}

